///////////////////////////////////
//Technik 2009
//Revolution Engine 0.5
///////////////////////////////////
//Main source file, contains the workflow of the engine

//header files
#include "REV.h"

//Internal structures
typedef struct _tracker
{
	OBJECT * target;
	struct _tracker * next;
	f32 z;
}TRACKER;

//Global variables
GXRModeObj *rMode;//GX render mode
void * frameBuffer[2] = { NULL, NULL};//To store external framebuffers, double buffering
u8 fb = 0;//Frame buffer index
Mtx view, modelM, modelview;
WPADData * wmData[4];
u16 w, h;
ROOT * mainRoot;//Root of the 3D space;
TRACKER * transQueue;
TRACKER * solidQueue;
TRACKER * zQueue;

//Private prototypes
void setUp3D(void);
void setUp2D(void);
void init3dConfig(void);
void init2dConfig(void);

//Public functions
void clasify3D(NODE * node)
{
	NODE * aux = node->child;
	while(aux)
	{
		clasify3D(aux);
		aux = aux->next;
	}
	if(node->type == NT_OBJECT)
	{
		TRACKER * auxT = (TRACKER*)malloc(sizeof(TRACKER));
		auxT->next = NULL;
		auxT->target = (OBJECT*)node;
		u8 tf = 0;//transparency flag
		if(auxT->target->clr.a != 255)
			tf = 1;
		else if(auxT->target->material) tf = 1;
		if(tf)
		{
			auxT->next = transQueue;
			transQueue = auxT;
		}
		else
		{
			auxT->next = solidQueue;
			solidQueue = auxT;
		}
	}
}

void orderQueue(CAMERA * cam)
{
	TRACKER * tz = transQueue;//Temporal storage Queue;
	TRACKER * aux, *t;
	transQueue = NULL;
	Vector v;
	aux = tz;
	while(aux)
	{
		tz = aux->next;
		aux->next = NULL;
		//
		v = aux->target->getPos();
		guVecMultiply(view, &v, &v);
		aux->z = v.z;
		//
		t = transQueue;
		if(t)
		{
			while(t->next)
			{
				if(aux->z > t->next->z)
					t = t->next;
				else break;
			}
			if(aux->z > t->z)
			{
				aux->next = t->next;
				t->next = aux;
			}
			else
			{
				aux->next = transQueue;
				transQueue = aux;
			}
		}
		else
		{
			transQueue = aux;
			aux->next = NULL;
		}
		//
		aux = tz;
	}
}

void setBGColor(GXColor clr)
{
	GX_SetCopyClear(clr, 0x00ffffff);
}

ROOT::ROOT()
{
	rootNode = NULL;
	rootN2D = NULL;
	views = NULL;
	fstLight = NULL;
	fstHTC = NULL;
	fstBtn = NULL;
}

void ROOT::init(f32 sx, f32 sy)
{
	rootNode = new NODE;
	rootN2D = new N2D;
	views = new VIEWPORT(vector3(0.0f, 0.0f, 0.01f), sx, sy, NULL);
}

void REV_init()
{
	//Local variables
	f32 yscale;
	u32 xfbHeight;
	////
	VIDEO_Init();//Video Library Init
	PAD_Init();//Gamecube Pad Library Init
	WPAD_Init();//Wiimote Library Init
	rMode = VIDEO_GetPreferredMode(NULL);//Set render mode
	// allocate 2 framebuffers for double buffering
	frameBuffer[0] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rMode));
	frameBuffer[1] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rMode));
	//Configure Video System
	VIDEO_Configure(rMode);
	VIDEO_SetNextFramebuffer(frameBuffer[fb]);//Initial Framebuffer
	VIDEO_Flush();
	VIDEO_WaitVSync();
	if(rMode->viTVMode&VI_NON_INTERLACE) VIDEO_WaitVSync();
	fb ^= 1;//Set next index
	//Set up the FIFO
	void *gp_fifo = memalign(32,DEFAULT_FIFO_SIZE);//Allocate
	memset(gp_fifo,0,DEFAULT_FIFO_SIZE);//and Clear
	//Init GX hardware
	GX_Init(gp_fifo,DEFAULT_FIFO_SIZE);
	// clears the bg to color and clears the z buffer
	GX_SetCopyClear(DEF_BG_COLOR , 0x00ffffff);
	//More initialization
	w = rMode->viWidth;
    h = rMode->viHeight;
	
	yscale = GX_GetYScaleFactor(rMode->efbHeight,rMode->xfbHeight);
	xfbHeight = GX_SetDispCopyYScale(yscale);
	GX_SetDispCopySrc(0,0,rMode->fbWidth,rMode->efbHeight);
	GX_SetDispCopyDst(rMode->fbWidth,xfbHeight);
	GX_SetFieldMode(rMode->field_rendering,((h==2*rMode->xfbHeight)?GX_ENABLE:GX_DISABLE));
	
	if (rMode->aa)
        GX_SetPixelFmt(GX_PF_RGB565_Z16, GX_ZC_LINEAR);
    else
        GX_SetPixelFmt(GX_PF_RGB8_Z24, GX_ZC_LINEAR);
	
	GX_CopyDisp(frameBuffer[fb],GX_TRUE);
	GX_SetDispCopyGamma(GX_GM_1_0);
	
	for(u8 i = 0; i < 4; i++)
	WPAD_SetVRes(i,w,h);
	
	//Hardware Initialized, now engine initialization
	//First create a Root
	mainRoot = new ROOT;
	mainRoot->init(w, h);
	defaultCam = new CAMERA;
	mainRoot->views->camera = defaultCam;
	mainRoot->views->sx = w;
	mainRoot->views->sy = h;
	GX_SetBlendMode(GX_BM_BLEND, GX_BL_SRCALPHA, GX_BL_INVSRCALPHA, GX_LO_CLEAR);
	GX_SetColorUpdate(GX_ENABLE);
	GX_SetAlphaUpdate(GX_ENABLE);
	GX_SetZCompLoc(GX_TRUE);
	GX_SetZMode (GX_TRUE, GX_LEQUAL, GX_TRUE);
	init3dConfig();
	init2dConfig();
	init2D(w, h);
	//Init filesystem
	fatInitDefault();
	REV_fontsInit();
}

void REV_preProcess()
{
	WPAD_ScanPads();
	PAD_ScanPads();
	//Update time
	updateTime();
	HTCAMERA * htc = mainRoot->fstHTC;
	for(u8 i = 0; i < 4; i++)
	{
		wmData[i] = WPAD_Data(i);
	}
	update2D();
	while(htc)//Update Head-Tracking cameras's point of view
	{
		htc->updateEye();
		htc = htc->nextHTC;
	}
}

void render(NODE * node, Vector camPos)
{
	if(node->flags & F_Visible)
	{
		GX_ClearVtxDesc();
		if(node->type == NT_OBJECT)
		{
			OBJECT * obj = (OBJECT*)node;
			if(obj->model)
			{
				u8 clrChannels = CC_DIFF, texChannels = 0;
				node->absMtx(modelM);
				guMtxConcat(view, modelM, modelview);
				GX_LoadPosMtxImm(modelview, GX_PNMTX0);
				
				if(obj->material)
				{
					obj->material->setTev(1, &clrChannels, &texChannels);			
					obj->model->render(modelM, camPos, clrChannels, texChannels, obj->material->specularity);
				}
				else
				{
					GX_SetVtxDesc(GX_VA_CLR0, GX_DIRECT);
					GX_SetNumChans(1);
					GX_SetTevColor(GX_TEVREG0,obj->clr);
					GX_SetNumTevStages(1);
					GX_SetTevColorIn(GX_TEVSTAGE0, GX_CC_ZERO, GX_CC_RASC, GX_CC_C0, GX_CC_ZERO);
					GX_SetTevAlphaIn(GX_TEVSTAGE0, GX_CA_ZERO, GX_CA_ZERO, GX_CA_ZERO, GX_CA_A0);
					GX_SetTevColorOp(GX_TEVSTAGE0, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
					GX_SetTevAlphaOp(GX_TEVSTAGE0, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
					GX_SetTevOrder(GX_TEVSTAGE0, GX_TEXCOORDNULL, GX_TEXMAP_NULL, GX_COLOR0A0);			
					obj->model->render(modelM, camPos, clrChannels, texChannels);
				}
			}
		}
	}
}

void REV_process()
{
	//Variables
	f32 preWait = 0, postWait = 0;
	static u8 firstFrame = 1;
	Mtx44 perspective;
	VIEWPORT * tView = mainRoot->views;
	TRACKER * auxT;
	setUp3D();
	//Wait just before drawing (instead of after), this should enhance performance
	VIDEO_Flush();
	preWait = (f32)(ticks_to_millisecs(gettime()));
 	VIDEO_WaitVSync();
	postWait = (f32)(ticks_to_millisecs(gettime()));
	GPUWaitTime = 0.001 * (postWait - preWait);
	//Clasify objects into solid or transparent queues
	//This is done before everything else because this clasification is the same for every viewport
	clasify3D(mainRoot->rootNode);
	//Render each Viewport into it's texture
	//GX_SetZMode (GX_TRUE, GX_LEQUAL, GX_TRUE);
	while(tView)
	{
		if(tView->flags & F_Visible)//Render only visible viewports
		{
			tView->setForRender(perspective);
			tView->camera->setForRender(view);
			//Before rendering the scene, render the skyBox
			GX_SetZMode (GX_FALSE, GX_LEQUAL, GX_TRUE);
			mainRoot->skyBox.render(tView->camera);
			GX_SetZMode (GX_TRUE, GX_LEQUAL, GX_TRUE);
			//Now render objects
			GX_SetCullMode(GX_CULL_BACK);
			GX_SetTevOp(GX_TEVSTAGE0,GX_PASSCLR);
			GX_SetTevOrder(GX_TEVSTAGE0, GX_TEXCOORDNULL, GX_TEXMAP_NULL, GX_COLOR0A0);
			GX_SetNumChans(1);
			auxT = solidQueue;
			while(auxT)
			{
				render(auxT->target, tView->camera->getPos());
				auxT = auxT->next;
			}
			orderQueue(tView->camera);
			auxT = transQueue;
			while(auxT)
			{
				render(auxT->target, tView->camera->getPos());
				auxT = auxT->next;
			}
			//Copy the embeded frame buffer to the texture
			tView->copyTexture();
		}
		tView = tView->nextVP;//prepare next viewport
	}
	while(solidQueue)
	{
		auxT = solidQueue;
		solidQueue = solidQueue->next;
		free(auxT);
	}
	while(transQueue)
	{
		auxT = transQueue;
		transQueue = transQueue->next;
		free(auxT);
	}
	//2D System
	//GX_SetZMode (GX_FALSE, GX_LEQUAL, GX_TRUE);
	GX_SetCopyFilter(rMode->aa,rMode->sample_pattern,GX_TRUE,rMode->vfilter);
	GX_SetViewport(0,0, w, h,0,1);
	GX_SetScissor(0,0, w, h);
	guOrtho(perspective,0,h,0,w, 1.0,100.0);
	GX_LoadProjectionMtx(perspective, GX_ORTHOGRAPHIC);
	GX_SetCullMode(GX_CULL_NONE);
	setUp2D();
	parse2D(mainRoot->rootN2D);
	order2D();
	render2D();
	GX_DrawDone();
	
	GX_CopyDisp(frameBuffer[fb],GX_TRUE);
	VIDEO_SetNextFramebuffer(frameBuffer[fb]);
	
	//Set out black screen after first frame
	if(firstFrame)
	{
		firstFrame = 0;
		VIDEO_SetBlack(FALSE);
	}
	
	fb ^= 1;
}

void REV_exit()
{
	WPAD_Flush(WPAD_CHAN_ALL);
	WPAD_Disconnect(WPAD_CHAN_ALL);
	WPAD_Shutdown();
	exit(0);
}

//Private functions
void setUp3D(void)
{
	GX_InvVtxCache();
	GX_InvalidateTexAll();
}

void setUp2D(void)
{
	GX_InvVtxCache();
	GX_InvalidateTexAll();
	GX_SetChanCtrl(GX_COLOR1A1,GX_FALSE,GX_SRC_VTX,GX_SRC_VTX,GX_LIGHTNULL,GX_DF_NONE,GX_AF_NONE);
	GX_SetChanCtrl(GX_COLOR0A0,GX_TRUE,GX_SRC_VTX,GX_SRC_VTX,GX_LIGHTNULL,GX_DF_NONE,GX_AF_NONE);
}

void init3dConfig(void)
{
	//Atributes format
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_POS, GX_POS_XYZ, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_NRM, GX_NRM_XYZ, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_CLR0, GX_CLR_RGBA, GX_RGBA8, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_CLR1, GX_CLR_RGBA, GX_RGBA8, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX0, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX1, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX2, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX3, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX4, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX5, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX6, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX7, GX_TEX_ST, GX_F32, 0);
}

void init2dConfig(void)
{
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_POS, GX_POS_XYZ, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_NRM, GX_NRM_XYZ, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_CLR0, GX_CLR_RGBA, GX_RGBA8, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_CLR1, GX_CLR_RGBA, GX_RGBA8, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_TEX0, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_TEX1, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_TEX2, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_TEX3, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_TEX4, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_TEX5, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_TEX6, GX_TEX_ST, GX_F32, 0);
	GX_SetVtxAttrFmt(GX_VTXFMT1, GX_VA_TEX7, GX_TEX_ST, GX_F32, 0);
}
