///////////////////////////////////
//Technik 2009
//Revolution Engine 0.5
///////////////////////////////////
//Materials source file

#include "REV.h"

MATERIAL::MATERIAL()
{
	flags = 0;
}

BMATERIAL::BMATERIAL(GXColor diffuse, GXColor specular, f32 specularity)
{
	type = MTT_BASIC;
	diffuseClr = diffuse;
	specularClr = specular;
	this->specularity = specularity;
	diffuseMap = NULL;
	specularMap = NULL;
}

void BMATERIAL::setTev(u8 nUVs, u8 * clrChannels, u8 * texChannels)
{
	u8 nTex = 0;
	GX_SetVtxDesc(GX_VA_CLR0, GX_DIRECT);//Diffuse lighting
	GX_SetVtxDesc(GX_VA_CLR1, GX_DIRECT);//Specular lighting
	*clrChannels = CC_DIFF + CC_SPEC;
	GX_SetNumChans(2);
	GX_SetNumTevStages(3);
	
	//TEVSTAGE 0
	if(diffuseMap && nUVs)//Diffuse mapping
	{
		GX_LoadTexObj(diffuseMap, GX_TEXMAP0);
		GX_SetTevColorIn(GX_TEVSTAGE0, GX_CC_ZERO, GX_CC_RASC, GX_CC_TEXC, GX_CC_ZERO);
		GX_SetTevAlphaIn(GX_TEVSTAGE0, GX_CA_ZERO, GX_CA_ZERO, GX_CA_ZERO, GX_CA_TEXA);
		nTex++;
	}
	else
	{
		GX_SetTevColor(GX_TEVREG0,diffuseClr);//Diffuse color
		GX_SetTevColorIn(GX_TEVSTAGE0, GX_CC_ZERO, GX_CC_RASC, GX_CC_C0, GX_CC_ZERO);
		GX_SetTevAlphaIn(GX_TEVSTAGE0, GX_CA_ZERO, GX_CA_ZERO, GX_CA_ZERO, GX_CA_A0);
	}
	GX_SetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD0, GX_TEXMAP0, GX_COLOR0A0);
	GX_SetTevColorOp(GX_TEVSTAGE0, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVREG2);
	GX_SetTevAlphaOp(GX_TEVSTAGE0, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVREG2);
	//TEVSTAGE 1
	if(specularMap && nUVs)
	{
		GX_LoadTexObj(specularMap, GX_TEXMAP1);
		GX_SetTevColorIn(GX_TEVSTAGE1, GX_CC_ZERO, GX_CC_RASC, GX_CC_TEXC, GX_CC_ZERO);
		GX_SetTevAlphaIn(GX_TEVSTAGE1, GX_CA_ZERO, GX_CA_RASA, GX_CA_TEXA, GX_CA_ZERO);
		GX_SetTevOrder(GX_TEVSTAGE1, GX_TEXCOORD0, GX_TEXMAP1, GX_COLOR1A1);
		nTex++;
	}
	else
	{
		GX_SetTevColor(GX_TEVREG1,specularClr);
		GX_SetTevColorIn(GX_TEVSTAGE1, GX_CC_ZERO, GX_CC_RASC, GX_CC_C1, GX_CC_ZERO);
		GX_SetTevAlphaIn(GX_TEVSTAGE1, GX_CA_ZERO, GX_CA_RASA, GX_CA_A1, GX_CA_ZERO);
		GX_SetTevOrder(GX_TEVSTAGE1, GX_TEXCOORDNULL, GX_TEXMAP_NULL, GX_COLOR1A1);
	}
	GX_SetTevColorOp(GX_TEVSTAGE1, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
	GX_SetTevAlphaOp(GX_TEVSTAGE1, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
	//TEVSTAGE 2
	GX_SetTevColorIn(GX_TEVSTAGE2, GX_CC_ZERO, GX_CC_CPREV, GX_CC_APREV, GX_CC_C2);
	GX_SetTevAlphaIn(GX_TEVSTAGE2, GX_CA_A2, GX_CA_ZERO, GX_CA_ZERO, GX_CA_APREV);
	GX_SetTevColorOp(GX_TEVSTAGE2, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
	GX_SetTevAlphaOp(GX_TEVSTAGE2, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
	GX_SetTevOrder(GX_TEVSTAGE2, GX_TEXCOORDNULL, GX_TEXMAP_NULL, GX_COLOR0A0);
	if(nTex)
	{
		GX_SetVtxDesc(GX_VA_TEX0, GX_INDEX16);
		*texChannels = 1;
	}
}

GXTexObj * loadPng(const char * filename)
{
	FILE * temp = fopen(filename, "rb");
	if(!temp)
		return NULL;
	else fclose(temp);
	GXTexObj * map = (GXTexObj*)memalign(32, sizeof(GXTexObj));
	PNGUPROP imgProp;
	IMGCTX ctx = PNGU_SelectImageFromDevice(filename);
	PNGU_GetImageProperties(ctx,&imgProp);
	void * data = memalign(32, imgProp.imgWidth * imgProp.imgHeight * 4);
	PNGU_DecodeTo4x4RGBA8 (ctx, imgProp.imgWidth, imgProp.imgHeight, data, 255);
	PNGU_ReleaseImageContext (ctx);
	DCFlushRange (data, imgProp.imgWidth * imgProp.imgHeight * 4);
	GX_InitTexObj (map, data, imgProp.imgWidth, imgProp.imgHeight, GX_TF_RGBA8, GX_CLAMP, GX_CLAMP, GX_FALSE);
	return map;//And return the texture
}
